/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */


#ifndef _LIGHT_KIS_
#define _LIGHT_KIS_

/****************************************************************/
/*********************** lightKIS Class *************************/
/****************************************************************/

template <class GENOME_TYPE>
class lightKIS
{
	private:
		/**
		 * Defines the coordinates of both segments in both genomes preserved by the object.
		 */
		GENOME_TYPE start1,start2,end1,end2;

		/**
		 * Orientation +/- .
		 */
		char Orientation;

		/**
		 * valid is an auxiliary variable which was used in the implementation. valid is initialized with the value 1.
		 */
		int valid;

		/**
		 * name is simply the label of the lightKIS object.
		 */
		string name;
	
	public:
	
		/**	
		 * Default constructor.
		 */
		lightKIS(){}
		
		/**
		 * Secondary constructor.
		 @param start11: The beginning coordinate of the segment of the first genome.
		 @param end12: The ending coordinate of the segment of the first genome.
		 @param start21: The beginning coordinate of the segment of the second genome.
		 @param end22: The ending coordinate of the segment of the second genome.
		 @param orientation: The orientation of the two segments.
		 */
		lightKIS(GENOME_TYPE start11,GENOME_TYPE end12,GENOME_TYPE start21,GENOME_TYPE end22,char orientation):start1(start11),start2(start21),end1(end12),end2(end22),Orientation(orientation),valid(1){}	
		
		/**
		 * Copy constructor.
		 * @param elm: lightKIS object const reference.
		 */ 
		lightKIS(const lightKIS<GENOME_TYPE>& elm);
	
		/**
		 * A method which labels the object with the given string.
		 * @param newName: string.
  		 */
		void nameIt(string& newName);
		
		/** 
		 * Returns the name of this object.
		 * @return string
		 */
		string getName()const;
		
		/**
		 * Returns the orientation of this object.
		 * @return char
		 */
		char get_Orientation()const;
		
		/**
		 * Assignment operator.
		 * Copies elm's fields to the current object fields.
		 * @param elm: const lightKIS<GENOME_TYPE>&
		 * @return lightKIS<GENOME_TYPE>&
		 */
		lightKIS<GENOME_TYPE>& operator=(const lightKIS<GENOME_TYPE>& elm); 

		/**
		 * Returns the details of both segments held by the calling object.
             * @param start11: A reference to the variable which will save the beginning value of the segment in the first genome.
           	  * @param end12: A reference to the variable which will save the ending value of the segment in the first genome.
	       * @param start21: A reference to the variable which will save the beginning value of the segment in the second genome.		
	     	  * @param end22: A reference to the variable which will save the ending value of the segment in the second genome.
		 */
		void get_Coords(GENOME_TYPE& start11,GENOME_TYPE& end12,GENOME_TYPE& start21,GENOME_TYPE& end22)const;

		/**
		 * Operator < . 
		 * Compares between this and elm according to start2 field.
		 * @param elm: const lightKIS<GENOME_TYPE>&
		 * @return bool
		 */
		bool operator<(const lightKIS<GENOME_TYPE>& elm)const;
		
		/**
		 * Operator > .
		 * Compares between this and elm according to start2 field.
		 * @param elm: const lightKIS<GENOME_TYPE>&
		 * @return bool
		 */
		bool operator>(const lightKIS<GENOME_TYPE>& elm)const;
		
		/**
		 * Operator == .
		 * Compares between this and elm according to coordinates fields (start1,start2,end1 and end2).
	  	 * @param elm: const lightKIS<GENOME_TYPE>&
		 * @return bool
 		 */
		bool operator==(const lightKIS<GENOME_TYPE>& elm)const;
		
		/**
		 * Updates the calling object with the given values.
		 * @param start1: The new beginning of the segment in the first genome.
		 * @param end1: The new ending of the segment in the first genome.
		 * @param start2: The new beginning of the segment in the second genome.
		 * @param end2: The new ending of the segment in the second genome.
		 */
		void update(GENOME_TYPE start1,GENOME_TYPE end1,GENOME_TYPE start2,GENOME_TYPE end2);
		
		/**
		 * Returns the valid variable value of the calling object.
		 */
		int getValid()const;
		
		/**
		 * Sets the value of the calling object valid variable to status.
		 * @param status: int
		 */
		void setValid(int status);
};


/*************************************************************************************/

template <class GENOME_TYPE>
void lightKIS<GENOME_TYPE>::update(GENOME_TYPE start1,GENOME_TYPE end1,GENOME_TYPE start2,GENOME_TYPE end2)
{
	this->start1=start1;	
	this->start2=start2;
	this->end1=end1;
	this->end2=end2;
}

/*************************************************************************************/

template <class GENOME_TYPE>
char lightKIS<GENOME_TYPE>::get_Orientation()const
{
	return Orientation;
}

/*************************************************************************************/	

template <class GENOME_TYPE>
void lightKIS<GENOME_TYPE>::nameIt(string& newName)
{
	name=newName;
}

/*************************************************************************************/

template <class GENOME_TYPE>
string lightKIS<GENOME_TYPE>::getName()const
{
	return name;
}

/*************************************************************************************/

template <class GENOME_TYPE>
bool lightKIS<GENOME_TYPE>::operator<(const lightKIS<GENOME_TYPE>& elm)const
{
	return (start2<elm.start2);
}

/*************************************************************************************/

template <class GENOME_TYPE>
bool lightKIS<GENOME_TYPE>::operator>(const lightKIS<GENOME_TYPE>& elm)const
{
	return (start2>elm.start2);
}

/*************************************************************************************/

template <class GENOME_TYPE>
bool lightKIS<GENOME_TYPE>::operator==(const lightKIS<GENOME_TYPE>& elm)const
{
	return ((start1==elm.start1)&&(start2==elm.start2)&&(end1==elm.end1)&&(end2==elm.end2));
}

/*************************************************************************************/

template <class GENOME_TYPE>
lightKIS<GENOME_TYPE>::lightKIS(const lightKIS<GENOME_TYPE>& elm)
{
		*this=elm;
}

/*************************************************************************************/

template <class GENOME_TYPE>
lightKIS<GENOME_TYPE>& lightKIS<GENOME_TYPE>::operator=(const lightKIS<GENOME_TYPE>& elm) 
{

	this->start1=elm.start1;
	this->start2=elm.start2;
	this->end1=elm.end1;
	this->end2=elm.end2;
	this->Orientation=elm.Orientation;
	this->name=elm.name;
	this->valid=elm.valid;

	return *this;
}

/*************************************************************************************/

template <class GENOME_TYPE>
void lightKIS<GENOME_TYPE>::get_Coords(GENOME_TYPE& start11,GENOME_TYPE& end12,GENOME_TYPE& start21,GENOME_TYPE& end22)const
{
	start11=start1;
	start21=start2;
	end12=end1;
	end22=end2;
}

/*************************************************************************************/

template <class GENOME_TYPE>
void lightKIS<GENOME_TYPE>::setValid(int status)
{
	if( (status<0)||(status>1) )
		return;
	valid=status;
}

/*************************************************************************************/

template <class GENOME_TYPE>
int lightKIS<GENOME_TYPE>::getValid()const
{
	return valid;
}

/*************************************************************************************/

#endif

