/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */

#include "cleanfile.h"
#include "seq.h"
#include "kises.h"
#include "prepro.h"
#include "globalVars.h"

int checkArgs(int argc,char** argv)
{
	int i,j;

	//errors.open(string(string(argv[argc-1])+string("error")).c_str());
	
	if((argc==2)&&(argv[1][0]=='-'))
	{
		if(argv[1][0]== '-')
		{
			string tmp="-help";
			if(tmp==string(argv[1]))
			{
				if(argc>2)
				{	
					cerr<<"It should be: pre-pro -help"<<endl;	
					//errors<<"It should be: pre-pro -help"<<endl;	
					return 1;
				}
				cerr<<"\n			***USAGE OF PRE-PRO***"<<endl<<endl;
				cerr<<"* pre-pro [options] path1/file1.anchor path2/file2.anchor path3/"<<endl<<endl;
				cerr<<"\t-help :"<<"\n\t\tDisplays this Help screen.\n"<<endl<<endl;
					cerr<<"\tgJL <long> :"<<"\n\t\tInitializes gapJoinLen variable with <long>\n\t\twhich is used for joining close gaps.\n\t\t(Default 110)\n"<<endl<<endl;
					cerr<<"\tgEL <long> :"<<"\n\t\tInitializes gapExtractLen variable with <long>\n\t\twhich is used for extracting long unmatched regions.\n\t\t(Default 200)\n"<<endl<<endl;
					cerr<<"\tsDiff <long> :"<<"\n\t\tInitializes stitchDifference variable with <long>\n\t\twhich is used for finding hits having similar distances.\n\t\t(Default 2000)\n"<<endl<<endl;
					cerr<<"\tsDis <long> :"<<"\n\t\tInitializes stitchDistance variable with <long>\n\t\twhich is used for finding close hits.\n\t\t(Default 15000)\n"<<endl<<endl;
					cerr<<"\tsMin <long> :"<<"\n\t\tInitializes stitchMinLen variable with <long>\n\t\twhich is used for removing short stitched hits.\n\t\t(Default 200)\n"<<endl<<endl;
					cerr<<"\t-circular <0/1><0/1> :"<<"\n\t\t\"-circular 00\" says that both genomes are not circular.\n\t\t(10-first genome is circular,01-2nd genome is \n\t\tcircular,11-both are circular, Default 00)"<<endl<<endl;
					cerr<<"\t-eValue <real> :"<<"\n\t\tDetermines the e-value threshold used for filtering\n\t\tlocal alignment hits.\n\t\tThe parameter is passed as is to YASS.\n\t\t(Default 0.01)"<<endl<<endl<<endl;
				cerr<<"*NOTE:"<<endl;	
					cerr<<"\tPRE-PRO assumes the existence of the files \"file1.genome\" and\n\t\"file2.genome\" under path1 and path2 respectively."<<endl<<endl;
					cerr<<"\tThe final result of PRE-PRO will be saved in the file \n\t\"path3/finalPREPRO.result\"."<<endl<<endl;
 				

				//errors<<"\n			***USAGE OF PRE-PRO***"<<endl<<endl;
				//errors<<"* pre-pro [options] path1/file1.anchor path2/file2.anchor path3/"<<endl<<endl;
				//errors<<"\t-help :"<<"\n\t\tDisplay this Help screen\n"<<endl<<endl;
				//	errors<<"\tgJL <long> :"<<"\n\t\tInitializes gapJoinLen variable with <long>\n\t\twhich is used for joining close gaps.\n\t\t(Default 110)\n"<<endl<<endl;
				//errors<<"\tgEL <long> :"<<"\n\t\tInitializes gapExtractLen variable with <long>\n\t\twhich is used for extracting long unmatched regions.\n\t\t(Default 200)\n"<<endl<<endl;
				//	errors<<"\tsDiff <long> :"<<"\n\t\tInitializes stitchDifference variable with <long>\n\t\twhich is used for finding hits having similar distances.\n\t\t(Default 2000)\n"<<endl<<endl;
				//	errors<<"\tsDis <long> :"<<"\n\t\tInitializes stitchDistance variable with <long>\n\t\twhich is used for finding close hits.\n\t\t(Default 15000)\n"<<endl<<endl;
				//	errors<<"\tsMin <long> :"<<"\n\t\tInitializes stitchMinLen variable with <long>\n\t\twhich is used for removing short stitched hits.\n\t\t(Default 200)\n"<<endl<<endl;
				//	errors<<"\t-circular <0/1><0/1> :"<<"\n\t\t\"-circular 00\" says that both genomes are not circular.\n\t\t(10-first genome is circular,01-2nd genome is \n\t\tcircular,11-both are circular, Default 00)"<<endl<<endl;
				//errors<<"\t-eValue <real> :"<<"\n\t\t"<<endl<<endl<<endl;
				//errors<<"*NOTE:"<<endl;	
				//	errors<<"\tPRE-PRO assumes the existence of the files \"file1.genome\" and\n\t\"file2.genome\" under path1 and path2 respectively."<<endl<<endl;
				//	errors<<"\tThe final result of PRE-PRO will be saved in the file \n\t\"path3/finalPREPRO.result\"."<<endl<<endl;
			}
			else
			{
				string tmp=string(argv[1]);
				cerr<<tmp<<" is an illegal option"<<endl;
				//errors<<tmp<<" is an illegal option"<<endl;
				return -1;
			}
			return 3;

		}
	}

	else
		if(argc<4)
		{
			cerr<<"Abnormal number of arguments"<<endl;
			//errors<<"Abnormal number of arguments"<<endl;
			return 4;
		}

	if(argc==4)
	{
		FILE* a,*b;
		if(((a=fopen(argv[1],"r"))==NULL)||((b=fopen(argv[2],"r"))==NULL))
		{
			cerr<<"Unable to open the files "<<argv[1]<<","<<argv[2]<<endl;
			//errors<<"Unable to open the files "<<argv[1]<<","<<argv[2]<<endl;
			return 5;
		}
		else
		{
			fclose(a);
			fclose(b);
		}
		
		input1=string(argv[1]);
		input2=string(argv[2]);
		input3=string(argv[3]);
		return 0;
	}

	for(i=1;i<argc-3;i++)
	
			if(argv[i][0]== 'g')
			{
				string tmp1("gJL"),tmp2("gEL");
				if(tmp1==string(argv[i]))
				{
					long gJL=0;
					int j=0;
					i++;
					while(argv[i][j]!='\0')
					{
						if( (argv[i][j]<'0')||(argv[i][j]>'9') )
						{
							cerr<<"Illegal gJL value"<<endl;
							//errors<<"Illegal gJL value"<<endl;
							return 6;
						}
						else
							gJL=gJL*10+argv[i][j]-'0';
						j++;
					}
						
					gapJoinLen=gJL;
				}
				else
				{
					if(tmp2==string(argv[i]))
					{
						long gEL=0;
						int j=0;
						i++;
						while(argv[i][j]!='\0')
						{
							if( (argv[i][j]<'0')||(argv[i][j]>'9') )
							{
								cerr<<"Illegal gEL value"<<endl;
								//errors<<"Illegal gEL value"<<endl;
								return 7;
							}
							else
								gEL=gEL*10+argv[i][j]-'0';
							j++;
						}
						
						gapExtractLen=gEL;
					}
					else
					{	
						string tmp=string(argv[i]);
						cerr<<tmp<<" is an illegal option"<<endl;
						//errors<<tmp<<" is an illegal option"<<endl;
						return -i;
					}
				}
			}
			
			else
				if(argv[i][0]== 's')
				{
					string tmp1="sDiff",tmp2="sDis",tmp3="sMin";
					if(tmp1==string(argv[i]))
					{
						long sDiff=0;
						int j=0;
						i++;
						while(argv[i][j]!='\0')
						{
							if( (argv[i][j]<'0')||(argv[i][j]>'9') )
							{
								cerr<<"Illegal sDiff value"<<endl;
								//errors<<"Illegal sDiff value"<<endl;
								return 8;
							}
							else
								sDiff=sDiff*10+argv[i][j]-'0';
							j++;
						}
						
						stitchDifference=sDiff;
					}
					else
					{
						if(tmp2==string(argv[i]))
						{
							long sDis=0;
							int j=0;
							i++;
							while(argv[i][j]!='\0')
							{
								if( (argv[i][j]<'0')||(argv[i][j]>'9') )
								{
									cerr<<"Illegal sDis value"<<endl;
									//errors<<"Illegal sDis value"<<endl;
									return 9;
								}
								else
									sDis=sDis*10+argv[i][j]-'0';
								j++;
							}
						
							stitchDistance=sDis;
						}
						else
						{	
							if(tmp3==string(argv[i]))
							{
								long sMin=0;
								int j=0;
								i++;
								while(argv[i][j]!='\0')
								{
									if( (argv[i][j]<'0')||(argv[i][j]>'9') )
									{
										cerr<<"Illegal sMin value"<<endl;
										//errors<<"Illegal sMin value"<<endl;
										return 10;
									}
									else
										sMin=sMin*10+argv[i][j]-'0';
									j++;
								}
						
								stitchMinLen=sMin;
							}
							else
							{
								string tmp=string(argv[i]);
								cerr<<tmp<<" is an illegal option"<<endl;
								//errors<<tmp<<" is an illegal option"<<endl;
								return -i;
							}
						}
					}
				}
		
				else
				{
					if((argv[i][0]== '-')&&(string(argv[i])==string("-help")))
					{
						cerr<<"It should be: pre-pro -help"<<endl;
						//errors<<"It should be: pre-pro -help"<<endl;	
						return 11;
					}
					else
						if((argv[i][0]== '-')&&(string(argv[i])==string("-circular")))
						{
							i++;
							
		if( (argv[i][0]!='\0')&&(argv[i][1]!='\0')&&((argv[i][0]=='0')||(argv[i][0]=='1'))&&((argv[i][1]=='0')||(argv[i][1]=='1')) &&(string(argv[i]).length()==2))
							{	
								
								circular1=argv[i][0]-'0';
								circular2=argv[i][1]-'0';
							}
							else
							{
								cerr<<"Illegal circular value"<<endl;
								//errors<<"Illegal circular value"<<endl;
								return 12;
							}
						}
						else
						{
							if((argv[i][0]== '-')&&(string(argv[i])==string("-eValue")))
							{
								i++;
								j=0;
								if((string(argv[i]).length()>0)&&(argv[i][0]!='.'))
								{									
									while(argv[i][j]!='\0')
									{
										if(((argv[i][j]>'9')||(argv[i][j]<'0'))&&(argv[i][j]!='.'))
											break;
										else
											j++;
									}
								
									if(argv[i][j]!='\0')
									{
										cerr<<"Illegal eValue value"<<endl;
										//errors<<"Illegal eValue value"<<endl;
										return 13;
									}
									else
										eValue=string(argv[i]);
								}
								else
								{	
									cerr<<"Illegal eValue value"<<endl;	
									//errors<<"Illegal eValue value"<<endl;
									return 13;
								}	
							}
							else
							{	
								string tmp=string(argv[i]);
								cerr<<tmp<<" is an illegal option"<<endl;
								//errors<<tmp<<" is an illegal option"<<endl;
								return -i;
							}
						}
				}
	FILE* a,*b;

	if(((a=fopen(argv[i],"r"))==NULL)||((b=fopen(argv[i+1],"r"))==NULL))
	{
		cerr<<"Unable to open the files "<<argv[i]<<","<<argv[i+1]<<endl;
		//errors<<"Unable to open the files "<<argv[i]<<","<<argv[i+1]<<endl;
		return 13;
	}
	else
	{
		fclose(a);
		fclose(b);
	}

	input1=string(argv[i]);
	input2=string(argv[i+1]);
	input3=string(argv[i+2]);

	return 0;
}

/******************************************************************************************************************/

int main(int argc,char** argv)
{
	/** Args check **/
	
	int argRes=checkArgs(argc,argv);

	if(argRes!=0)
		return argRes;

	unsigned int i=0;
	string buff;

	/** File cleaning **/

	vector< SEQ<long> > org1,org2;
	vector< SEQ<long> >::iterator s,e,s1,e1;
	KISES<long> findKises;
	
	if(input3[input3.length()-1]!='/')
		input3+=string("/");

	i=input1.length();

	while((i>0)&&(input1[i]!='/'))
		i--;

	if((i==0)&&(input1[0]!='/'))
		buff=input1;
	else
		buff=input1.substr(i+1,input1.length());
	
	buff = input3+buff+string(".clean");

	CleanFile<long> in1(input1);
	in1.MakeLegal();
	in1.getCleanRes(buff);

	i=input2.length();

	while((i>0)&&(input2[i]!='/'))
		i--;

	if((i==0)&&(input2[0]!='/'))
		buff=input2;
	else
		buff=input2.substr(i+1,input2.length());
	
	buff = input3+buff+string(".clean");

	CleanFile<long> in2(input2);
	in2.MakeLegal();
	in2.getCleanRes(buff);
	
	/** Finding intersection **/

	buff=input3+string("Intersection.res");

	in1.intersection(in2,org1,org2,buff);

	/** Finding KISs **/

	string first(input1.erase(input1.length()-6)+"genome"),scnd(input2.erase(input2.length()-6)+"genome");
	PREPRO<long> prepro(first,scnd);
	long len;

	buff=input3+string("KISs.res");

	if(org1.size()==0)
	{
		cerr<<"No intersection found between the given files."<<endl;
		exit(1);
	}

	vector< KIS<long> > KISES=findKises.find_KISes(org1,org2,buff.c_str());
	vector< lightKIS<long> > newKISes;
	string* GAres=NULL;

	/** Performing global alignment on KISs **/

	if( KISES.size() > 0 )	
	{
		GAres=prepro.gAlign(KISES,len);

	/** Extracting gaps **/

		if( GAres != NULL )
			newKISes=prepro.extractGaps(GAres,KISES);
	}
	else
	{
		cerr<<"KIS table is empty!"<<endl;
		exit(1);
	}

	/** Extracting breakpoints **/

	string* chunks=prepro.extractBP(prepro.extractBPCoords(newKISes));
	if(chunks==NULL)
	{
		cerr<<"No breakpoints were found."<<endl;
		exit(1);
	}

	/** Performing local alignment **/
	cerr<<"Performing local alignment ...\n"<<endl;
	vector< lightKIS<long> > res1=prepro.lAlign(chunks[0],1),res2=prepro.lAlign(chunks[1],0);

	/** Stitching Hits **/
	cerr<<"Stitching Hits ... \n"<<endl;
	vector< lightKIS<long> > reso1=prepro.stitchHits(res1),reso2=prepro.stitchHits(res2);

	long len2 = reso1.size()+1;
	long len1=reso2.size();

	for(int i=0;i<len1;i++)
	{
		string buff;
		prepro.toString(len2++,buff);
		buff=string("SIS-")+buff;
		reso2[i].nameIt(buff);
		reso1.push_back(reso2[i]);
	}

	vector< lightKIS<long> > gaps;

	/** Performing global alignment on Hits **/
	/** Extracting gaps from Hits **/

	gaps=prepro.extractGapsFromHits(prepro.GAOnStitchedHits(reso1),reso1);

	/** Adding stitched Hits to the KIS table **/

	prepro.addToKISTable(gaps,newKISes);

	buff=input3+string("finalPREPRO.result");

	prepro.printToFile(newKISes,buff);

	return 0;
}
