/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */



#ifndef _MATCH_
#define _MATCH_

/*************************************************************/
/*********************** match Class *************************/
/*************************************************************/

template <class T>
class match
{
	private:

		/**
		 * The element held by match class.
	  	 */
		T elm;
		
		/**
		 * An index used for mapping elm to another T element held in an indexed data structure (see kisses.h).
		 */
		int mtch;

	public:
		
		/**
		 * Sets the index of the mapping element.
		 * @param m: An integer defining an index.
		 */
		void set_match(int m);
		
		/**
		 * Operator < .
		 * Compares between this and elem.
		 * @param elem: const match& (The comparative match object).
		 * @return bool: true if (*this)<elem, false otherwise.
		 */
		bool operator<(const match& elem)const;
		
		/**
		 * Returns the mapping index.	
		 * @return int: The mapping index.
		 */
		int get_match()const;
		
		/**
		 * Copies elm to the current object.
		 * @param elm: A reference to a T object.
		 * @return A reference to the current object.
		 */
		match<T>& add(const T& elm);
		
		/**
		 * Returns a reference to the stored T object.
		 * @return A reference to the T object.
		 */
		const T& get_elm()const;
};

/*************************************************************/

template <class T>
void match<T>::set_match(int m)
{
	mtch=m;
}

/*************************************************************/

template <class T>
int match<T>::get_match()const
{
	return mtch;
}

/*************************************************************/

template <class T>
bool match<T>::operator<(const match& elem)const
{
	return ( elm < elem.elm );
}

/*************************************************************/

template <class T>
match<T>& match<T>::add(const T& elm)
{
	this->elm=elm;
	return (*this);
}

/*************************************************************/

template <class T>
const T& match<T>::get_elm()const
{
	return elm;
}

/*************************************************************/

#endif

