/*
 *  YASS 1.10
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>


/* 1) include utils macro */
#include "util.h"
/* 2) include global variables */
#include "global_var.h"
/* 3) the current file defs */
#include "display.h"
/* 4) other files */
#include "align.h"
#include "prdyn.h"
#include "tuple.h"
#include "proba.h"
#include "kword.h"

int stat1[3] = { 0, 0, 0 };
int stat2[3] = { 0, 0, 0 };
int transindels[3] = { 0, 0, 0 };

LAresult Display_Alignements(int querysize,
			char *dataquery,
			char *dataqueryrev, char *queryname,
			char *datatext,
			int nbchunks, char **chunkname, int *chunksize,
			int *chunkstart)
{

/***************************************************************************/

    LAresult res=(LAresult)malloc(sizeof(struct LAres)),tmp,node;
    if(res==NULL)
    {
	fprintf(stderr,"Memory Allocation Error.\n");
	exit(1); 
    }

    res->next=NULL; 	

/***************************************************************************/
    int nb_displayed_alignments = 0;
    MA *ma = NULL;

    /* [0] header if needed */
/*    if (gp_display == 2) {
	fprintf(OUTSTREAM,
		"# Fields: Query id, Subject id, %% identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score\n");
    }
    if (gp_display == 3) {
	fprintf(OUTSTREAM,
		"# q. start,\tq. end,\ts. start,\ts. end,\tq. length,\ts.length,\tf/r\tbit score\te-value\n");
    }*/


    /* [1] Effective display of each Memorized Alignment */
    for (ma = gv_first_MA;
	 ma != NULL && nb_displayed_alignments < gp_nbmaxlines;
	 ma = ma->next) {

	nb_displayed_alignments++;
/*
#ifdef DEBUG
	DisplayMA(ma);
#endif
*/
	switch (gp_display) {
	case 0:
	case 1:
	    /*  (1.1) alignment position repeat */
	    if (!(ma->reverse))
		fprintf(OUTSTREAM, "*(%d-%d)(%d-%d)"  ,
			BCOUNT + ma->left_pos_begin   ,
			BCOUNT + ma->left_pos_end  - 1,
			BCOUNT + ma->right_pos_begin  ,
			BCOUNT + ma->right_pos_end - 1);

	    else {
		fprintf(OUTSTREAM, "*(%d-%d)(%d-%d)",
			BCOUNT + querysize - ma->left_pos_begin - 1,
			BCOUNT + querysize - ma->left_pos_end      ,
			BCOUNT + ma->right_pos_begin               ,
			BCOUNT + ma->right_pos_end              - 1);
	    }
		/* (1.2) print details (size, evalue, etc ...) */
		fprintf(OUTSTREAM, " Ev: %g s: %d/%d %c\n",
			Evalue(gp_k_blast, gp_lambda_blast, gp_querysize,
			       gp_textsize, ma->blastscore),
			ma->left_pos_end - ma->left_pos_begin   ,
			ma->right_pos_end - ma->right_pos_begin ,
			(ma->reverse) ? 'r' : 'f');
	    
	    if (gp_display == 0)
		break;


	    /* (2.1) print chunk names (reverse/complement) */
	    fprintf(OUTSTREAM, "* %s %s / %s (forward strand)\n",
		    queryname,
		    (ma->
		     reverse) ? "(reverse complementary strand)" :
		    "(forward strand)", chunkname[ma->i_chunk]
		);
	    /* (2.2) score bitscore and statistics */
	    fprintf(OUTSTREAM, "* score = %d : bitscore = %.2f\n",
		    (ma->blastscore), (float) BitScore(gp_k_blast,
						       gp_lambda_blast,
						       (ma->blastscore)));



	    memset(stat1, '\0', 3 * sizeof(int));
	    memset(stat2, '\0', 3 * sizeof(int));
	    memset(transindels, '\0', 3 * sizeof(int));

	    alignment_SG_stats_on_MA((ma->
				      reverse) ? dataqueryrev : dataquery,
				      datatext + chunkstart[ma->i_chunk]
				     , ma, transindels, stat1, stat2,0);

	    fprintf(OUTSTREAM,
		    "* mutations per triplet %d, %d, %d (%.2e) | ts : %d tv : %d\n\n",
		    stat1[0], stat1[1], stat1[2],
		    (float) (P_mutation_bias(stat1)), transindels[0],
		    transindels[1]
		);
	    /* (2.3) display alignment */
	    display_alignment_SG_on_MA((ma->reverse) ? dataqueryrev :
				       dataquery,
				       datatext + chunkstart[ma->i_chunk]
				       , ma);
	    
	    /* DisplayMA(ma); */
	    
	    break;		/* case 1: */

	case 2:

		 /* (3) blast like tabular format */
	    /* (3.1) fasta name */

	   /*************************************************************************/		
		
		node = (LAresult)malloc(sizeof(struct LAres));
	    	if(node==NULL)
	    	{
			fprintf(stderr,"Memory Allocation Error.\n");
			exit(1); 
    		}
		node->next=NULL;

		node->name=(char*)malloc(strlen(chunkname[ma->i_chunk])+1);
		if(node->name==NULL)
	    	{
			fprintf(stderr,"Memory Allocation Error.\n");
			exit(1); 
    		}

		strcpy(node->name,chunkname[ma->i_chunk]);

	/**************************************************************************/
 
	  /*  fprintf(OUTSTREAM, "%s\t%s\t", queryname,
	  	    chunkname[ma->i_chunk]);*/
	    /* (3.2) stats analysis */
	   /* memset(stat1, '\0', 3 * sizeof(int));
	    memset(stat2, '\0', 3 * sizeof(int));
	    memset(transindels, '\0', 3 * sizeof(int));

	    alignment_SG_stats_on_MA((ma->
				      reverse) ? dataqueryrev : dataquery,
				     datatext + chunkstart[ma->i_chunk]
				     , ma, transindels, stat1, stat2,0);
	   */ /* (3.3) percent identity */
	    /*fprintf(OUTSTREAM, "%3.2f\t",
		    100 -
		    100.000 *
		    ((double)
		     (transindels[0] + transindels[1] +
		      transindels[2]) / MAX((ma->left_pos_end -
					     ma->left_pos_begin ),
					    (ma->right_pos_end -
					     ma->right_pos_begin ))));*/
	    /* (3.4) size */
	/*    fprintf(OUTSTREAM, "%d\t",
		    MAX((ma->left_pos_end - ma->left_pos_begin),
			(ma->right_pos_end - ma->right_pos_begin)));*/
	    /* (3.5) mutations indels */
	//    fprintf(OUTSTREAM, "%d\t%d\t", transindels[0] + transindels[1], transindels[2]);

/******************************************************************************************************/
		if(!(ma->reverse))
		{
			node->start1=BCOUNT + ma->left_pos_begin;
			node->end1=BCOUNT + ma->left_pos_end - 1;
			node->start2=BCOUNT + ma->right_pos_begin;
			node->end2=BCOUNT + ma->right_pos_end - 1;
		}
		else
		{	
			node->start1=BCOUNT + querysize-1 - ma->left_pos_end;
			node->end1=BCOUNT + querysize-1 - ma->left_pos_begin - 1;
			node->start2=BCOUNT + ma->right_pos_end;
			node->end2=BCOUNT + ma->right_pos_begin - 1;
		}
		if(ma->reverse)
			node->orientation='-';
		else
			node->orientation='+';

		tmp=res->next;
		res->next=node;
		node->next=tmp;

/******************************************************************************************************/ 

	    /* (3.6) alignment position repeat */
	  /*  if (!(ma->reverse))
		fprintf(OUTSTREAM, "%d\t%d\t%d\t%d\t",
			BCOUNT + ma->left_pos_begin,
			BCOUNT + ma->left_pos_end - 1,
			BCOUNT + ma->right_pos_begin,
			BCOUNT + ma->right_pos_end - 1);
	    else
		fprintf(OUTSTREAM, "%d\t%d\t%d\t%d\t", 
			BCOUNT + querysize - ma->left_pos_end      ,  
			BCOUNT + querysize - ma->left_pos_begin - 1,  
			BCOUNT + ma->right_pos_end - 1,	             
			BCOUNT + ma->right_pos_begin );
	    fprintf(OUTSTREAM, "%.2g\t%.3g\n",
		    Evalue(gp_k_blast, gp_lambda_blast, gp_querysize,
			   gp_textsize, ma->blastscore),
		    BitScore(gp_k_blast, gp_lambda_blast,
			     ma->blastscore));*/
	    break;		/* case 2: */
	case 3:
	    /*  (1.1) alignment position repeat */
	    if (!(ma->reverse))
		fprintf(OUTSTREAM, "%d\t%d\t%d\t%d",
			BCOUNT + ma->left_pos_begin,
			BCOUNT + ma->left_pos_end - 1,
			BCOUNT + ma->right_pos_begin,
			BCOUNT + ma->right_pos_end - 1);
	    else
		fprintf(OUTSTREAM, "%d\t%d\t%d\t%d",
			BCOUNT + querysize - ma->left_pos_end,       /* inverted pair */
			BCOUNT + querysize - ma->left_pos_begin - 1,
			BCOUNT + ma->right_pos_begin,                /* normal pair */
			BCOUNT + ma->right_pos_end              - 1);

	    /* (1.2) print details (size, f/b , bitscore , evalue ) */
	    fprintf(OUTSTREAM, "\t%d\t%d\t%c\t%2g\t%3g\n",
		    ma->left_pos_end  - ma->left_pos_begin,
		    ma->right_pos_end - ma->right_pos_begin,
		    (ma->reverse) ? 'r' : 'f',
		    BitScore(gp_k_blast, gp_lambda_blast, ma->blastscore),
		    Evalue(gp_k_blast, gp_lambda_blast, gp_querysize, gp_textsize, ma->blastscore)
		    );
	    break;

	}			/* switch */
    }				/* for */
    return res;
}




int Display_Stats(Feature  feature[2])
{
    fprintf(stderr, "* Statistical parameters\n");
    fprintf(stderr, "   lambda = %g\n", gp_lambda_blast);
    fprintf(stderr, "   K      = %g\n", gp_k_blast);
    fprintf(stderr, "   rho    = %d\n", gp_rho_stat);
    fprintf(stderr, "   delta  = %d\n", gp_delta_stat);

#ifdef STATS
    fprintf(stderr, "   ----\n");
    /* preprocess */
    fprintf(stderr, "   # keys removed     =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_keys_removed,
	    feature[1].nb_keys_removed,
	    feature[0].nb_keys_removed + feature[1].nb_keys_removed);
    /* hits */
    fprintf(stderr, "   -\n");
    fprintf(stderr, "   # single hit       =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_seeds, 
	    feature[1].nb_seeds, 
	    feature[0].nb_seeds + feature[1].nb_seeds);
    fprintf(stderr, "   # chains built     =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_chains_built , 
	    feature[1].nb_chains_built ,
	    feature[0].nb_chains_built + feature[1].nb_chains_built
	    );
    fprintf(stderr, "   # align. tested    =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_chains_tested ,
	    feature[1].nb_chains_tested ,
	    feature[0].nb_chains_tested + feature[1].nb_chains_tested
	    );
    fprintf(stderr, "   # align. detected  =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_ma , 
	    feature[1].nb_ma , 
	    feature[0].nb_ma + feature[1].nb_ma
	    );
    /* regroup */
    fprintf(stderr, "   -\n");
    fprintf(stderr, "   # grouping tests   =   f:%9d   r:%9d   total:%9d \n",
	    feature[0].nb_postprocessed_grouping_tests,
	    feature[1].nb_postprocessed_grouping_tests,
	    feature[0].nb_postprocessed_grouping_tests + feature[1].nb_postprocessed_grouping_tests
	    );
    fprintf(stderr, "   # align. linked    =   f:%9d   r:%9d   total:%9d \n",
	    feature[0].nb_postprocessed_grouping_links,
	    feature[1].nb_postprocessed_grouping_links,
	    feature[0].nb_postprocessed_grouping_links + feature[1].nb_postprocessed_grouping_links
	    );
    fprintf(stderr, "   # align. postpros  =   f:%9d   r:%9d   total:%9d \n",
	    feature[0].nb_postprocessed_ma,
	    feature[1].nb_postprocessed_ma,
	    feature[0].nb_postprocessed_ma + feature[1].nb_postprocessed_ma
	    );

    /* cpu-times */
    fprintf(stderr, "   ----\n");
     fprintf(stderr, " pre-process cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_pre  /CLOCKS_PER_SEC,
	    feature[1].clock_pre  /CLOCKS_PER_SEC,
	    (feature[0].clock_pre   + feature[1].clock_pre  )/CLOCKS_PER_SEC,
	    (int)(
	    (feature[0].clock_pre   + feature[1].clock_pre  )/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post  + 1e-6)*100)
	    );
    fprintf(stderr, "    channing cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_chain/CLOCKS_PER_SEC,
	    feature[1].clock_chain/CLOCKS_PER_SEC,
	    (feature[0].clock_chain + feature[1].clock_chain)/CLOCKS_PER_SEC,
	    (int)(
	    (feature[0].clock_chain + feature[1].clock_chain)/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post  + 1e-6)*100)
	    );
    fprintf(stderr, "   alignment cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_align/CLOCKS_PER_SEC,
	    feature[1].clock_align/CLOCKS_PER_SEC,
	    (feature[0].clock_align + feature[1].clock_align)/CLOCKS_PER_SEC,
	     (int)(
	    (feature[0].clock_align + feature[1].clock_align)/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post   + 1e-6)*100)
	    );
    fprintf(stderr, "post-process cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_post/CLOCKS_PER_SEC,
	    feature[1].clock_post/CLOCKS_PER_SEC,
	    (feature[0].clock_post + feature[1].clock_post)/CLOCKS_PER_SEC,
	     (int)(
	    (feature[0].clock_post  + feature[1].clock_post )/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post   + 1e-6)*100)
	    );    


 /*
    
    if (clock_align + clock_chain + clock_post == 0)
      clock_align = clock_chain = clock_post = 1;

    fprintf(stderr, "       chaining  time = %d s [ %d%%]\n",
	    (int) (0.5 +
		   gv_time_spent * (clock_chain /
				 (clock_align + clock_chain + clock_post)))
	    ,
	    (int) (0.5 +
		   100 * (clock_chain /
			  (clock_align + clock_chain + clock_post)))
	);
    fprintf(stderr, "       alignment time = %d s [ %d%%]\n",
	    (int) (0.5 +
		   gv_time_spent * (clock_align /
				 (clock_align + clock_chain + clock_post)))
	    ,
	    (int) (0.5 +
		   100 * (clock_align /
			  (clock_align + clock_chain + clock_post)))
	);
    fprintf(stderr, " post-processing time = %d s [ %d%%]\n",
	    (int) (0.5 +
		   gv_time_spent * (clock_post /
				 (clock_align + clock_chain + clock_post)))
	    ,
	    (int) (0.5 +
		   100 * (clock_post /
			  (clock_align + clock_chain + clock_post)))
	);
    */

    fprintf(stderr, " real-time spent = %ds\n", gv_time_spent);

#endif	

    return 0;

}




int Display_Params()
{
    int i;
    fprintf(stderr, "\n");
    fprintf(stderr, RULE);

    /* filenames */
    if (gp_nbfiles <= 1) {
	fprintf(stderr, "  DNA file : ");
	fprintf(stderr, "\"%s\", %d chunk(s), selected chunk: %d (size : %d bp)\n", gp_files[0], gp_nbchunks_query, gp_selection_fasta, gp_querysize_chunk);
    } else {
	fprintf(stderr, "  first DNA file : ");
	fprintf(stderr, "\"%s\", %d chunk(s), selected chunk: %d (size : %d bp)\n", gp_files[0],  gp_nbchunks_query, gp_selection_fasta, gp_querysize_chunk);
	fprintf(stderr, "  second DNA file : ");
	fprintf(stderr, "\"%s\"  %d chunk(s), (overall size : %d bp) \n", gp_files[1], gp_nbchunks_text, gp_textsize);
    }

    fprintf(stderr, "* Parameters\n");
    for (i = 0; i < gp_nb_seeds; i++)
	fprintf(stderr,
		"  seed[%d]: pattern = \"%s\", weight = %d bits, span = %d bases\n",
		i, gp_motifs[i], gp_seeds_bitweight[i], gp_seeds_span[i]
		);
    fprintf(stderr,
	    "  cost (match,mismatch,transition,other) = %d,%d,%d,%d\n",
	    gp_costs[0], gp_costs[1], gp_costs[2], gp_costs[3]
	    );
    fprintf(stderr, 
	    "  cost gaps (opening,continous) = %d,%d\n",
	    gp_cost_gap_opened, gp_cost_gap_continued
	    );
    fprintf(stderr, 
	    "  mutation rate = %d%%\n", 
	    gp_mutations_percent
	    );
    fprintf(stderr, 
	    "  indels rate = %d%%\n",
	    gp_indels_percent
	    );
    fprintf(stderr, 
	    "  tolerance  = %d%%\n", 
	    gp_alpha_percent);
    fprintf(stderr, 
	    "  minimum score seeds = %d\n",
	    gp_minimum_score_seeds
	    );
    fprintf(stderr,
	    "  minimal score = %d\n",
	    gp_minscore
	    );
    return 0;
}




/* Dot progress bar : 
 * "launch and lets take a coffee"(TM)  progress bar
 */



int Display_Progress(unsigned int pos, unsigned int maxpos,Feature  *feature)
{
    if (gp_reverse == 2) {
	    if (feature->last_point * maxpos < pos * (NBL / 2)) {
		while (feature->last_point * maxpos < pos * (NBL / 2)
		       && feature->last_point < NBL) {
		    fprintf(stderr, ".");
		    fflush(NULL);
		    (feature->last_point)++;
		}
	    }
    } else {
	if (feature->last_point * maxpos < pos * (NBL)) {
	    while (feature->last_point * maxpos < pos * (NBL) && feature->last_point < NBL) {
		fprintf(stderr, ".");
		fflush(NULL);
		(feature->last_point)++;
	    }
	}
    }
    return 0;
}


