/*
 *  YASS 1.10
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

/* 1) include utils macro */
#include "util.h"
/* 2) include global variables */
#include "global_var.h"
/* 3) the current file defs */
#include "kword.h"




/*
 *   Build a  DNA bases table  (code 0..3) from a DNA file
 * 
 * - Input  : "filename"                                    : DNA file name
 * - Output : "p_data"                                      : return the allocated data table pointer 
 *            "p_datasize"                                  : return the allocated data table size 
 *            "p_nbchunks","p_chunkname", and "p_chunksize" : return respectively the number of chunks fond in the multifasta file, their names and their size
 *                                                            (set to NULL if you do not need these informations
 */

/*****************************************************************************************************/


int CreateData( /*in */ char *filename, /*out */ char **p_data, /*out */
	       int *p_datasize, /*out */ int *p_nbchunks, /*out */
	       char ***p_chunkname, /*out */ int **p_chunksize, /*out */
		int **p_chunkstart, /* out */ int nb_letters[4],char *seq )
{
    int i = 0, j = 0,l=0;
    int car = 0;
    char * p_car;
    int datasize = 0;
    int cumulength = 0;
    char *data = NULL;
    FILE *file = NULL;
    long k=0;

    int nbchunks = 0;
    char **chunkname = NULL;
    int *chunksize = NULL;
    int *chunkstart = NULL;

    /* (1) count number of DNA bases availaible */

/*
 *      A --> adenosine           M --> A C (amino)
 *      C --> cytidine            S --> G C (strong)
 *      G --> guanine             W --> A T (weak)
 *      T --> thymidine           B --> G T C
 *      U --> uridine             D --> G A T
 *      R --> G A (purine)        H --> A C T
 *      Y --> T C (pyrimidine)    V --> G C A
 *      K --> G T (keto)          N --> A G C T (any)
 */

if(!seq){
    if (!(file = fopen(filename, "r"))) {
      fprintf(stderr,"* Error : unable to open \"%s\": errno = %i \n", filename,errno);
      exit(-1);
    }
}

if(seq){
    
    while (seq[k]!='\0') {
      car = seq[k++];
      
      if (car > (char)127){
	fprintf(stderr,"* Error : non ASCII character at position %i\n", datasize);
	exit(-1);
      }
	
      switch (car) {
	case 'A':
	case 'a':
	case 'T':
	case 't':
	case 'G':
	case 'g':
	case 'C':
	case 'c':
	case 'U':
	case 'u':
	case 'R':
	case 'r':
	case 'Y':
	case 'y':
	case 'K':
	case 'k':
	case 'M':
	case 'm':
	case 'S':
	case 's':
	case 'W':
	case 'w':
	case 'B':
	case 'b':
	case 'D':
	case 'd':
	case 'H':
	case 'h':
	case 'V':
	case 'v':
	case 'N':
	case 'n':
	    datasize++;
	    break;
	case ' ':
	case '\t':
	case '\n':
	    break;

	case '>':
	    nbchunks++;
	    do {
		/* seek eol */
		car = seq[k++];
	    } while (car != '\n');
	    break;
	default:
	  
	    /* while ((car = fgetc(file)) != '\n'); */
	    break;
	}			/*switch */
    }				/*while */
}

else{
    /* comment */
    while ((car = fgetc(file)) != EOF) {
      if (car == EOF)
	break;
      
      if (car > (char)127){
	fprintf(stderr,"* Error : non ASCII character at position %i\n", datasize);
	exit(-1);
      }
	
      switch (car) {
	case 'A':
	case 'a':
	case 'T':
	case 't':
	case 'G':
	case 'g':
	case 'C':
	case 'c':
	case 'U':
	case 'u':
	case 'R':
	case 'r':
	case 'Y':
	case 'y':
	case 'K':
	case 'k':
	case 'M':
	case 'm':
	case 'S':
	case 's':
	case 'W':
	case 'w':
	case 'B':
	case 'b':
	case 'D':
	case 'd':
	case 'H':
	case 'h':
	case 'V':
	case 'v':
	case 'N':
	case 'n':
	    datasize++;
	    break;
	case ' ':
	case '\t':
	case '\n':
	    break;

	case '>':
	    nbchunks++;
	    do {
		/* seek eol */
		car = fgetc(file);
	    } while (car != '\n' && car != EOF);
	    break;
	default:
	  
	    /* while ((car = fgetc(file)) != '\n'); */
	    break;
	}			/*switch */
    }				/*while */
}

    if (!datasize) {
      fprintf(stderr,"* Error : empty file \"%s\" ",filename);
      exit(-1);
    }
if(!seq)
    fclose(file);

    /* (2) allocate one memory blok for the DNA data */

    data = (char *) MALLOC((unsigned) datasize*sizeof(char));
    ASSERT(data, CreateData);


    if (nbchunks > 0) {
	chunkname = (char **) MALLOC((unsigned) nbchunks* sizeof(char *));
	ASSERT(chunkname, CreateData);
	chunksize = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunksize, Createdata);
	chunkstart = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunkstart, Createdata);
    }



    /* (3) fill the DNA data */
if(seq){
    k=0;
    /* description of sentance */
    while (seq[k]) {
	car=seq[k++];
	switch (car) {
	case 'A':
	case 'a':
	    data[i++] = (char) 0;
	    nb_letters[0]++;
	    break;
	case 'T':
	case 't':
	    data[i++] = (char) 1;
	    nb_letters[1]++;
	    break;
	case 'G':
	case 'g':
	    data[i++] = (char) 2;
	    nb_letters[2]++;
	    break;
	case 'C':
	case 'c':
	    data[i++] = (char) 3;
	    nb_letters[3]++;
	    break;
	case 'Y':
	case 'y':
	    data[i++] = (char) 4;
	    break;
	case 'S':
	case 's':
	    data[i++] = (char) 5;
	    break;
	case 'U':
	case 'u':
	    data[i++] = (char) 6;
	    break;
	case 'N':
	case 'n':
	    data[i++] = (char) 7;
	    break;
	case 'R':
	case 'r':
	    data[i++] = (char) 8;
	    break;
	case 'M':
	case 'm':
	    data[i++] = (char) 9;
	    break;
	case 'W':
	case 'w':
	    data[i++] = (char) 10;
	    break;
	case 'D':
	case 'd':
	    data[i++] = (char) 11;
	    break;
	case 'H':
	case 'h':
	    data[i++] = (char) 12;
	    break;
	case 'V':
	case 'v':
	    data[i++] = (char) 13;
	    break;
	case 'K':
	case 'k':
	    data[i++] = (char) 14;
	    break;
	case 'B':
	case 'b':
	    data[i++] = (char) 15;
	    break;
	case ' ':
	case '\t':
	case '\n':
	    break;
	case '>':
	    /* --[A] chunk name / size update */
	    if (j > 0) {
		chunksize[j - 1] = i - cumulength;
	    }
	    cumulength = i;
	    chunkstart[j] = i;
	    chunkname[j] = (char *) MALLOC((unsigned) 256 * sizeof(char));
	    while(seq[k]!='\n')
		chunkname[j][l++]=seq[k++];
	    (chunkname[j])[l] = '\0';
          l=0;
	    /* remove '\n' */
	    p_car = chunkname[j];
	    do {
	      if (*p_car == '\n'){
		*p_car = '\0';
		break;
	      }
	      p_car++;
	    } while(*p_car != '\0');
	    j++;
	    break;
	default:
	  /* ignore */;
	} /* switch */
    } /* while */
}
else{
    if (!(file = fopen(filename, "r")))
	return -3;
    /* description of sentance */
    while ((car = fgetc(file)) != EOF) {
	if (car == EOF)
	    break;
	switch (car) {
	case 'A':
	case 'a':
	    data[i++] = (char) 0;
	    nb_letters[0]++;
	    break;
	case 'T':
	case 't':
	    data[i++] = (char) 1;
	    nb_letters[1]++;
	    break;
	case 'G':
	case 'g':
	    data[i++] = (char) 2;
	    nb_letters[2]++;
	    break;
	case 'C':
	case 'c':
	    data[i++] = (char) 3;
	    nb_letters[3]++;
	    break;
	case 'Y':
	case 'y':
	    data[i++] = (char) 4;
	    break;
	case 'S':
	case 's':
	    data[i++] = (char) 5;
	    break;
	case 'U':
	case 'u':
	    data[i++] = (char) 6;
	    break;
	case 'N':
	case 'n':
	    data[i++] = (char) 7;
	    break;
	case 'R':
	case 'r':
	    data[i++] = (char) 8;
	    break;
	case 'M':
	case 'm':
	    data[i++] = (char) 9;
	    break;
	case 'W':
	case 'w':
	    data[i++] = (char) 10;
	    break;
	case 'D':
	case 'd':
	    data[i++] = (char) 11;
	    break;
	case 'H':
	case 'h':
	    data[i++] = (char) 12;
	    break;
	case 'V':
	case 'v':
	    data[i++] = (char) 13;
	    break;
	case 'K':
	case 'k':
	    data[i++] = (char) 14;
	    break;
	case 'B':
	case 'b':
	    data[i++] = (char) 15;
	    break;
	case ' ':
	case '\t':
	case '\n':
	    break;
	case '>':
	    /* --[A] chunk name / size update */
	    if (j > 0) {
		chunksize[j - 1] = i - cumulength;
	    }
	    cumulength = i;
	    chunkstart[j] = i;
	    chunkname[j] = (char *) MALLOC((unsigned) 256 * sizeof(char));
	    fgets(chunkname[j], 256, file);
	    (chunkname[j])[256 - 1] = '\0';
	    /* remove '\n' */
	    p_car = chunkname[j];
	    do {
	      if (*p_car == '\n'){
		*p_car = '\0';
		break;
	      }
	      p_car++;
	    } while(*p_car != '\0');
	    j++;
	    break;
	default:
	  /* ignore */;
	} /* switch */
    } /* while */
    fclose(file);
}

    /* --[B] last chunk size updated */
    if (j > 0) {
      chunksize[j - 1] = i - cumulength;
    }

    /* --[C] Raw format : simulate one big fasta chunk */
    if (nbchunks == 0) {
	nbchunks  = 1;
	chunkname = (char **) MALLOC((unsigned) nbchunks* sizeof(char *));
	ASSERT(chunkname, CreateData);
	chunksize = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunksize, Createdata);
	chunkstart = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunkstart, Createdata);
	chunkname[0] = filename;
	chunksize[0] = datasize;
	chunkstart[0] = 0;
    }

#ifdef DEBUG_CREATEDATA
    for (i = 0; i < nbchunks; i++)
	fprintf(stderr,
		" chunksize = %i chunkstart = %i chunkname = \"%s\" \n",
		chunksize[i], chunkstart[i], chunkname[i]);
#endif

    /* (4) return allocated and computed data */
    *p_datasize = datasize;
    *p_data = data;
    if (p_nbchunks != NULL)
	*p_nbchunks = nbchunks;
    if (p_chunkname != NULL)
	*p_chunkname = chunkname;
    if (p_chunksize != NULL)
	*p_chunksize = chunksize;
    if (p_chunkstart != NULL)
	*p_chunkstart = chunkstart;
    return 0;
}


/*****************************************************************************************************/


/*
 *   Build a DNA bases of the complementry reverse straind 
 * 
 * - Input  : "data_in"     : DNA array (returned by CreateData for example)
 *            "datasize"    : DNA array size
 * - Output : "p_data_out"  : DNA array containing the complementary reverse straind
 *            
 */


int CreateReverseComplement( /*in */ char *data_in, /*in */ int datasize,
			    /*out */ char **p_data_out)
{
    int i;
    char *data = (char *) MALLOC((unsigned) datasize* sizeof(char));
    ASSERT(data, CreateReverseComplement);

    for (i = 0; i < datasize; i++)
      data[datasize - 1 - i] = (char)COMPLEMENT(data_in[i]);
    
    *p_data_out = data;
    return 0;
}




/*
 * seed pattern size computation 
 * (gp_seed_bitweight, gp_seed_span)
 */


int ComputeLength()
{
    int seed;
    int i;
    gp_seeds_span_max = 0;
    gp_seeds_span_min = INFINITY_INT;
    for (seed = 0; seed < gp_nb_seeds; seed++) {
	gp_seeds_bitweight[seed] = 0;
	gp_seeds_span[seed] = 0;

	for (i = 0; i < (int) strlen(gp_motifs[seed]); i++) {
	    if (gp_motifs[seed][i] == '#')
		(gp_seeds_bitweight[seed]) += 2;
	    else if (gp_motifs[seed][i] == '@')
		(gp_seeds_bitweight[seed]) += 1;
	    else
	      /* if another char is given : replace it by _*/
	      gp_motifs[seed][i] = '_';
	    gp_seeds_span[seed]++;
	}
	gp_motifs[seed][gp_seeds_span[seed]] = '\0';
	gp_seeds_span_max = MAX(gp_seeds_span_max,gp_seeds_span[seed]);
	gp_seeds_span_min = MIN(gp_seeds_span_min,gp_seeds_span[seed]);
    }
    return 0;
}




void AutomataOverlaps(int delph)
{
    int i, j;
    int p_delph = 1 << delph;

    gp_automata = (int *) MALLOC(sizeof(int) * p_delph);
    ASSERT(gp_automata, AutomataOverlaps);
    memset(gp_automata, 0, sizeof(int) * p_delph);

    for (i = 0; i < p_delph; i++) {
	j = i;
	while (j != 0) {
	    j = j & (j - 1);
	    gp_automata[i]++;
	}
    }
}


static void BuildRedMask(int seed, int * p_table) {  
  int  i,j;
  char seed_element = 0;
  for ( j = 0 ; j < 4 ; j++ ) {
    p_table[j] = 0;
    for ( i = 0 ; i < gp_seeds_span[seed] ; i++ ) {
      seed_element = gp_motifs[seed][i];
      switch (seed_element) {
      case '#' :
	p_table[j] <<= 2;
	p_table[j] |= ((int) j & 0x3);
	break;
      case '@' :
	p_table[j] <<= 1;
	p_table[j] |= ((int) j & 0x1);
	break;
      }
    }
  }
}


/*
 * Create a linkedlist of kwords with param "list" and index "first"
 *
 * - Input : "data"    : dna input table 
 *           "datasize": dna input table size
 *           "seed"    : seed number
 * - Output: "p_first" : return a table of size (2^w). For each code w, it contains the first 
 *                       position occurence of the seed with such code on the sequence
 *           "p_list"  : return a table of size (keysize). For each position i, it contains next
 *                       position j such that seed i and j code are equal. 
 */


#define ISTART 0
#define PASSREDMASK(w,rtable) (((w) != (rtable[0])) && \
                               ((w) != (rtable[1])) && \
                               ((w) != (rtable[2])) && \
                               ((w) != (rtable[3])))
int CreateKeyList( /* in */ char *data,     /* in  */ int datasize,    /* in */ int seed,
		   /* out */ int **p_first, /* out */ int * p_first_size, 
		   /* out */ int **p_list,  /* out */ int * p_list_size, /* out */ Feature * feature)
{


  int *bucklet = NULL;
  int *first = NULL;
  int *list = NULL;
  int *count = NULL;
#ifdef KEYLISTCOMPRESS
  int *listcompress = NULL;
#endif
  int sizeofbucklet = 1;
  int keysize = datasize - gp_seeds_span[seed] + 1;
    
  int redmask[4];
  
  /* prune algorithm */
  int maxchainsize = 0;
  int nbkeys       = 0;
  int fastcountall = 0;
  int fastcounter[sizeof(unsigned int) * 8] = {0};
  int imask      [sizeof(unsigned int) * 8] = {0};
  
  int  i=0, j=0, w=0;

  /* (1) allocation of 3 table */
  sizeofbucklet <<= (gp_seeds_bitweight[seed]);
  
  bucklet = (int *) MALLOC((unsigned) sizeofbucklet * sizeof(int));
  ASSERT(bucklet, CreateKeyList);
  memset(bucklet, 0xff, (unsigned) sizeofbucklet * sizeof(int));
  
  first = (int *) MALLOC((unsigned) (sizeofbucklet+1) * sizeof(int));
  ASSERT(first, CreateKeyList);
  memset(first, 0xff, (unsigned) sizeofbucklet * sizeof(int));
  
  count = (int *) MALLOC((unsigned) sizeofbucklet * sizeof(int));
  ASSERT(count, CreateKeyList);
  memset(count, 0x0, (unsigned) sizeofbucklet * sizeof(int));
  
  list  = (int *) MALLOC((unsigned) keysize * sizeof(int));
  ASSERT(list, CreateKeyList);
  memset(list, 0xff, (unsigned) keysize * sizeof(int));
  
  /* (2) list creation algorithm */
  BuildRedMask(seed,redmask); /* redmask is unalowed key = "0, all G, all C, all A, all T" */
  for (j = 0; j < keysize; j++) {
    KEY(w,data,j,seed);
    if (PASSREDMASK(w,redmask)) {
      if ( first[w] == -1 ) {
	first[w] = j;
      } else {
	list[bucklet[w]] = j;
      }
      bucklet[w] = j;
      count[w]++;
      nbkeys++;
      maxchainsize = MAX(count[w], maxchainsize);
    }
  }
  FREE(bucklet,(unsigned) sizeofbucklet * sizeof(int));
  
  
  /* (3) median-mean selection */
  
  /* (3.0) init counters */
  imask[0] = 1;
  for (i = 1 ; i < sizeof(int)*8 ; i++ ) {
    imask[i] = ((imask[i-1]) << 1) | 1 ;
    fastcounter[i] = 0;
  }
  
  /* (3.1) fast approximate counting */ 
  for (w = 0; w < sizeofbucklet; w++) {
    for (i = ISTART ; i < sizeof(int)*8 ; i++ ){
      if ( ( imask[i]  & count[w] ) == count[w]) {
	fastcounter[i]++;
	break;
      }
    }
  }

#ifdef DEBUGFASTCOUNT
  for  (i = ISTART ; i < sizeof(int)*8 ; i++ ) {
    fprintf(stderr, " fastcount[%d] = %d \n", i , fastcounter[i]);
  }
#endif

  /* (3.2) select part to be removed part */
  for (i = ISTART ; i < sizeof(int)*8 ; i++ ) {
    fastcountall += fastcounter[i];
    if (fastcountall * 1.0 > sizeofbucklet * (1.0 - (gp_t + 1e-32)/sqrt(sizeofbucklet)))
      break;
  }

  /* (3.3) remove gp_t percent */
  for (w = 0; w < sizeofbucklet; w++) {
    if ((count[w] & imask[i]) != count[w]) {
      first[w] = -1;
      nbkeys -= count[w];
#ifdef STATS	
      feature->nb_keys_removed++;
#endif
#ifdef DEBUGRMKEY
      fprintf(stderr,"- removing key : w=%x, nbocc=%d\n",w,count[w]); 
#endif
    }
  }
  
  FREE(count,(unsigned) sizeofbucklet * sizeof(int));
  
#ifdef KEYLISTCOMPRESS
  listcompress =  (int *) MALLOC((unsigned) (nbkeys + 1) * sizeof(int));
  ASSERT(listcompress, CreateKeyList);
  i = 0;
  for (w = 0; w < sizeofbucklet; w++) {
    j = first[w];
    first[w] = i;
    while (j != -1) {
      listcompress[i++] = j;
      j = list[j];
    }
  }
  first[sizeofbucklet] = i;
#endif

  /* 4) list "copy" */
#ifdef KEYLISTCOMPRESS
  FREE(list, keysize * sizeof(int));
  *p_list       = listcompress;
  *p_list_size  = (nbkeys  + 1 ) * sizeof(int);
#else
  *p_list       = list;
  *p_list_size  = keysize * sizeof(int);
#endif
  *p_first      = first;
  *p_first_size = (sizeofbucklet + 1) * sizeof(int);
  return 0;

}




/*
 * COMPLEXITY CHECK
 */
double Entropy(char *data, int pos, int len, int k)
{
    int p4k = 1 << (k << 1);	/*4^k */
    int mask = 0;
    int key = 0;
    int i = 0;
    double result = 0.0;
    int *freq = (int *) MALLOC(p4k * sizeof(int));
    ASSERT(freq, Entropy);
    for (i = 0; i < p4k; i++)
	freq[i] = 0;
    
    for (i = 0; i < k; i++) {
	mask <<= 2;
	mask |= 3;
	key <<= 2;
	key |= (int) data[pos + i];
    }


    for (i = k; i < len; i++) {
	key <<= 2;
	key |= (int) data[pos + i];
	key &= mask;
	freq[key]++;
    }



    for (i = 0; i < p4k; i++) {
	if (freq[i] != 0) {
	    result -=
		((double) (freq[i]) / (double) (len - k + 1)) *
		log((double) (freq[i]) / (double) (len - k + 1));
	}
    }
    result /= log(2.0000);
    FREE(freq,p4k * sizeof(int));
    return result;
}




/*
 * Display sequence fragments
 */

int DisplaySequence(char *data, int pos, int len) {
  
    int j;
    for (j = pos; j < pos + len; j++) {
      fprintf(stderr,"%c", lookup[(int) data[j]]);
    }
    fprintf(stderr,"\n");
    return 0;
}


int DisplayData(char * filename, char * data, int size, int nbchunks, char ** chunkname, int * chunksize, int * chunkstrt) {
  int i;
  fprintf(stderr,"- filename : \t \"%s\"\n",filename);
  fprintf(stderr,"- size     : \t %d\n",size);
  fprintf(stderr,"- nbchunks : \t %d\n",nbchunks);
  for( i = 0 ; i < nbchunks ; i++ ) {
    fprintf(stderr,"> chunkname : \"%s\",  chunkstart: %d,   chunksize: %d\n",chunkname[i],chunkstrt[i],chunksize[i]);
    DisplaySequence(data,chunkstrt[i],chunksize[i]);
  }
  return 0;
}

int DisplaySeeds(){
  int i = 0;
  for( i = 0 ; i < gp_nb_seeds ; i++){
    fprintf(stderr,"- gp_motif[%d] = \"%s\", span = %3d , bitweight = %3d\n",i,gp_motifs[i],gp_seeds_span[i],gp_seeds_bitweight[i]);
  }
  return 0;
}
