/*
 *  YASS 1.10
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include "global_var.h"
#include "tuple.h"
#include "prdyn.h"
#include "proba.h"
#include "kword.h"


#ifdef INLINE
inline
#endif
tuple * CreateTuple(int occurrence, int diagonal, int seed)
{

    tuple *t = (tuple *) MALLOC((unsigned) sizeof(tuple));
    ASSERT(t, CreateTuple);
    t->occurrence = occurrence;
    t->diagonal = diagonal;
    t->leftsize = 0;
    t->bitweight = gp_seeds_bitweight[seed];
    t->next = NULL;
    t->lastSeed = seed;
    t->overlappedhits = 0x00000001;
    return t;
}


#ifdef INLINE
inline
#endif
tuplelist * CreateTupleList()
{
  tuplelist *tl = (tuplelist *) MALLOC((unsigned) sizeof(tuplelist));
  ASSERT(tl, CreateTupleList);
  tl->first_tuple = NULL;
  tl->next = NULL;
  return tl;
}



#ifdef INLINE
inline
#endif
int KeepTupleList(tuplelist * tl, tuple * tfirst, tuple * tlast, MA * ma)
{
    tuple *t = tl->first_tuple->next;
    tuple *t_prev = tl->first_tuple;
    if (tfirst == tl->first_tuple) {
	tl->first_tuple = tlast->next;
    } else {
	while (t != tfirst) {
	    t_prev = t;
	    t = t->next;
	}
	t_prev->next = tlast->next;
    }
    ma->first_tuple = tfirst;
    tlast->next = NULL;

    return 0;
}



/* only used in "Align_Single" and "Align_Double" */

#ifdef INLINE
inline
#endif
void FreeTupleList(tuplelist ** p_tl, tuplelist ** p_tl_prev, tuplelist **last_tl)
{
    tuple *t, *t_prev;

    if (*last_tl == *(p_tl))
	*last_tl = *(p_tl_prev);

    t = (*(p_tl))->first_tuple;
    while (t != NULL) {
	t_prev = t;
	t = t->next;
	FREE(t_prev,sizeof(tuple));
    }

    (*(p_tl_prev))->next = (*(p_tl))->next;
    FREE(*(p_tl),sizeof(tuplelist));

    *(p_tl) = (*(p_tl_prev))->next;
}


#ifdef INLINE
inline
#endif
    MA * CreateMA(int left_pos_begin, int right_pos_begin,
		  int left_pos_end, int right_pos_end,
		  tuplelist * tuple_list,
		  tuple * first_tuple_aligned,
		  tuple * last_tuple_aligned, int blastscore,
		  Feature * feature
		  )
{

    MA *ma = (MA *) MALLOC((unsigned) sizeof(MA));
    ASSERT(ma, memorize_alignment);
    
    /* 1) initialise ma */
    ma->next = NULL;
    ma->left_pos_begin = left_pos_begin;
    ma->right_pos_begin = right_pos_begin;
    ma->left_pos_end = left_pos_end;
    ma->right_pos_end = right_pos_end;
    ma->blastscore = blastscore;
    
    /* 2) keep the seeds chain inside MA */
    KeepTupleList(tuple_list, first_tuple_aligned, last_tuple_aligned, ma);

    ma->reverse = (char) feature->reverse;
    ma->i_chunk = feature->i_chunk;
    
    /* 3) add  "ma" in the MA_linked_list */
    if (feature->first_MA == NULL) {
	feature->first_MA = ma;
    } else {
	feature->last_MA->next = ma;
    } 
    feature->last_MA = ma;
   
#ifdef STATS	
    feature->nb_ma++;
#endif
    

    /* left_correction take effect on TL_POS(t): we can do alignments before calling the "left_correction"
     * function on each MA, while left_correction global variable is != 0, or either after the "left_correction()"
     * call provided that left_correction global variable IS = 0.
     */

    /* correction due to the algorithm "Assemble_Double" over "Assemble_Single" */
    
    LeftCorrection_MA(ma,feature->left_correction);

    return ma;
}

#ifdef INLINE
inline
#endif
void FreeMA(MA * ma)
{
    tuple *t_, *t_prev_;
    t_ = ma->first_tuple;
    while (t_) {
	t_prev_ = t_;
	t_ = t_->next;
	FREE(t_prev_,sizeof(tuple));
    }
    FREE(ma,sizeof(MA));
}

#ifdef INLINE
inline
#endif
int LeftCorrection_MA(MA * ma,int left_correction)
{
    tuple *t = ma->first_tuple;
    while (t) {
	t->diagonal -= left_correction;
	t = t->next;
    }
    return 0;
}

#define QS_CRITERION blastscore

int QuickSort_MAList(MA ** p_firstMA,
		     MA * last_ma /* set to NULL at the begining */ )
{

    MA *ma_pivot = *(p_firstMA);
    MA *ma_list = ma_pivot;
    MA *ma_current;
    MA *ma_prev_current = ma_pivot;
    int valeur_pivot;
    int valeur;

    if (ma_pivot == last_ma)
	return 0;

    ma_current = ma_pivot->next;
    valeur_pivot = ma_pivot->QS_CRITERION;

    if (ma_current == last_ma)
	return 0;


    do {

	valeur = ma_current->QS_CRITERION;
	if (valeur > valeur_pivot) {
	    /* put in the head of the list */
	    ma_prev_current->next = ma_current->next;
	    ma_current->next = ma_list;
	    ma_list = ma_current;
	    /* MAJ */
	    ma_current = ma_prev_current->next;
	} else {
	    /* MAJ */
	    ma_current = ma_current->next;
	    ma_prev_current = ma_prev_current->next;
	}
    } while (ma_current != last_ma);


    QuickSort_MAList(&(ma_list), ma_pivot);	/* qsort first part */
    QuickSort_MAList(&(ma_pivot->next), ma_current);	/* qsort second part */

    /* set the begining of the list at its firtst element */
    *(p_firstMA) = ma_list;

    return 0;
}

/*
 * ListSort
 */

int ListSort_MAList(MA ** p_firstMA, MA ** p_lastMA /* var set to NULL */ ,
		    int mask /* 0x40000000 */ )
{

    MA *ma_0 = NULL, *ma_1 = NULL;
    MA *ma_0_top = NULL, *ma_1_top = NULL;

    MA *_ma = *(p_firstMA);

    while (_ma) {
	if ((_ma->QS_CRITERION) & mask) {
	    if (!ma_1)
		ma_1 = _ma;
	    else
		ma_1_top->next = _ma;
	    ma_1_top = _ma;
	} else {
	    if (!ma_0)
		ma_0 = _ma;
	    else
		ma_0_top->next = _ma;
	    ma_0_top = _ma;
	}
	_ma = _ma->next;
    }

    mask >>= 1;

    if (ma_0_top) {
	ma_0_top->next = NULL;
	if (mask > 0)
	    ListSort_MAList(&ma_0, &ma_0_top, mask);
    }
    if (ma_1_top) {
	ma_1_top->next = NULL;
	if (mask > 0)
	    ListSort_MAList(&ma_1, &ma_1_top, mask);
    }


    /* chain the two sublists into one */
    if (ma_1_top) {
	*(p_firstMA) = ma_1;
	if (ma_0_top) {
	    ma_1_top->next = ma_0;
	    *(p_lastMA) = ma_0_top;
	} else {
	    *(p_lastMA) = ma_1_top;
	}
    } else {
	*(p_firstMA) = ma_0;
	*(p_lastMA) = ma_0_top;
    }

    return 0;
}


/* 
 * Remove MA with Evalue over MaxEvalue 
 */

int Prune_MAList(MA ** p_firstMA, double MaxEvalue)
{
    MA *ma, *ma_prev;
    int minScore =
	MinScore(gp_k_blast, gp_lambda_blast, gp_querysize, gp_textsize,
		 MaxEvalue);

    ma = *(p_firstMA);
    if (!ma)
	return 0;

    /* find the first element OK and set it to *(p_firstMA) */
    while (ma->blastscore < minScore) {
	ma_prev = ma;
	*(p_firstMA) = ma = ma->next;
	FreeMA(ma_prev);
	if (!ma)
	    return 0;
    }

    ma_prev = ma;
    ma = ma_prev->next;

    /* continue with the others MAs */
    while (ma) {
	if (ma->blastscore < minScore) {
	    ma_prev->next = ma->next;
	    FreeMA(ma);
	    ma = ma_prev->next;
	} else {
	    ma_prev = ma;
	    ma = ma->next;
	}
    }

    return 0;
}

int EntropyFilter_MA(MA * ma, char *dataquery,  char *datachunktext)
{
    int k_entropy = 3;
    return
	Entropy(dataquery,
		ma->left_pos_begin,
		ma->left_pos_end - ma->left_pos_begin,
		k_entropy) < gp_entropy_min
	||
	Entropy(datachunktext,
		ma->right_pos_begin,
		ma->right_pos_end - ma->right_pos_begin,
		k_entropy) < gp_entropy_min;

}



/*
 * Entropy filter : remove low complexity regions
 */

int EntropyFilter_MAList(Feature * f, char *dataquery, 
			 char *datatext, int *datatext_start)
{
    MA * ma = NULL, * ma_prev = NULL;

    ma = f->first_MA;

    /* (0) clean-up if nothing */
    if (!ma) {
      f->last_MA = NULL;
      return 0;
    }

    /* (1) find the first element OK and set it to *(p_firstMA) */
    while (EntropyFilter_MA (ma, dataquery, datatext + datatext_start[ma->i_chunk])) {
      MA * to_free_ma = ma;
      ma = ma->next;
      FreeMA(to_free_ma);
      if (!ma) {
	f->first_MA = f->last_MA = NULL;
	return 0;
      }
    }    
    f->first_MA = f->last_MA = ma;    

    ma_prev = ma;
    ma = ma_prev->next;

    /* continue with the others MAs */
    while (ma) {
	if ( EntropyFilter_MA(ma, dataquery, datatext + datatext_start[ma->i_chunk])) {
	  MA * to_free_ma = ma;
	  ma = ma->next;
	  ma_prev->next = ma;
	  FreeMA(to_free_ma);
	} else {
	  ma_prev = ma;
	  ma = ma->next;
	}
    }
    f->last_MA = ma_prev;
    return 0;
}




/*
 * Display for debuging effort
 */


int DisplayTuple(tuple * tuple)
{
  int left_correction = 0;
  printf("\t [(%9d,%9d),(%9d,%9d)] \t diag : %9d \t size : %9d \n", TBL_POS(tuple), TEL_POS(tuple), TBR_POS(tuple), TER_POS(tuple), tuple->diagonal , TSIZE(tuple));
  return 0;
}


int DisplayListTupleList(tuplelist * tuplelist)
{
  while (tuplelist) {
    DisplayTupleList(tuplelist);
    tuplelist = tuplelist->next;
  }
  return 0;
}


int DisplayTupleList(tuplelist * tuplelist)
{
  tuple *tuple;
  if (tuplelist->first_tuple) {
    tuple = tuplelist->first_tuple;
    while (tuple) {
      DisplayTuple(tuple);
      tuple = tuple->next;
    }
    printf("\n");
  }
  return 0;
}

int DisplayGap(tuple * t1, tuple * t2)
{
  printf("\t intergaps {%d,%d}\n",TGAP_L(t1,t2),TGAP_R(t1,t2));
  assert(TGAP_L(t1,t2) >= 0);
  assert(TGAP_R(t1,t2) >= 0);
  /* if (TGAP_L(t1,t2)*TGAP_R(t1,t2) > MAX_WINDOW_REGROUP)
   *   printf("\t FIXME \n");
   */
  return 0;
}

int DisplayMA(MA * ma)
{
  int left_correction = 0;
  
  tuple * tuple = ma->first_tuple;
 
  while (tuple) {
    printf("\t");DisplayTuple(tuple);
    assert(TBL_POS(tuple)   >= ma->left_pos_begin);
    assert(TEL_POS(tuple)   <= ma->left_pos_end);
    assert(TBR_POS(tuple)   >= ma->right_pos_begin);
    assert(TBR_POS(tuple)   <= ma->right_pos_end);
    if (tuple->next){
      printf("\t");DisplayGap(tuple,tuple->next);
    }
    tuple = tuple->next;
  }
  printf("\t << {(%d,%d), (%d,%d)}\n", ma->left_pos_begin, ma->left_pos_end, ma->right_pos_begin, ma->right_pos_end);
  printf("\n");
  return 0;
}

int DisplayListMA (MA * ma){

  while (ma){
    printf("\t MA : {(%d,%d),(%d,%d)}\n", ma->left_pos_begin, ma->left_pos_end, ma->right_pos_begin, ma->right_pos_end);
    
    /* ma attributes */
    assert(ma->reverse  <= 1);
    assert(ma->i_chunk  >= 0);
    assert(ma->i_chunk  < gp_nbchunks_text);
    assert (ma->blastscore >= 0);
    
    /* ma properties */
    assert(ma->left_pos_begin   >= 0);
    assert(ma->left_pos_begin   <= ma->left_pos_end);
    assert(ma->left_pos_end     <=  gp_querysize_chunk);
    assert(ma->right_pos_begin  >= 0);
    assert(ma->right_pos_begin  <= ma->right_pos_end);
    assert(ma->right_pos_end    <= gp_chunksize_text[ma->i_chunk]);
     
    /* ma tuples */ 
    DisplayMA(ma);
    
    ma = ma -> next;
  }

  return 0;
}
