/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */



#include "umr.h"
#include "globalVars.h"
#include <fstream>


/**************************************************************************************************/
/*********************** Extracting unmatched regions auxiliary functions *************************/
/**************************************************************************************************/

long extractUnmatchedRegions(string seq1,string seq2,long** gaps)
{	
	long len1,len2;
	
	long** gaps1= allocateMtx(seq1.length(),2);
	long** gaps2= allocateMtx(seq2.length(),2);
	
	if( (seq1.length()==0) || (seq2.length()==0) )
	{
		cerr<<"Input error to function extractUnmatchedRegions !"<<endl;
		//errors<<"Input error to function extractUnmatchedRegions !"<<endl;
		//finish.open((input3+string("done")).c_str());
		exit(1);
	}

	len1=findGaps(seq1,gaps1);
	len2=findGaps(seq2,gaps2);

	len1=joinGaps(gaps1,len1);
	len2=joinGaps(gaps2,len2);
		
	len1=extractGaps(gaps1,len1);
	len2=extractGaps(gaps2,len2);
	
	getIntersection(gaps,gaps1,gaps2,len1,len2);
	
	freeMtx(gaps1,seq1.length());
	freeMtx(gaps2,seq2.length());
	
	return len1+len2;

}

/************************************************************************************/

long findGaps(string seq,long** gaps)
{
	unsigned long i=0,indx=0;

	for(;i<seq.length();i++)
	{	
		if(seq[i]=='-')
		{
			gaps[indx][0]=i;
			while(seq[i]=='-')i++;
			gaps[indx++][1]=i-1;
		}
	}
	
	return indx;
}

/************************************************************************************/

long joinGaps(long** gaps,long len)
{
	long i,indx=0;
	node* tmp=new node[len];	
	
	for(i=0;i<len;i++)
	{
		tmp[i].start=gaps[i][0];
		tmp[i].end=gaps[i][1];
		tmp[i].valid=1;
	}
	
	for(i=1;i<len;i++)
	{
		if( (unsigned long)(tmp[i].start-tmp[i-1].end-1) <= gapJoinLen)
		{
			tmp[i].start=tmp[i-1].start;
			tmp[i-1].valid=0;
		}
	}
	
	for(i=0;indx<len;indx++)
	{
		if(tmp[indx].valid)
		{
			gaps[i][0]=tmp[indx].start;	
			gaps[i++][1]=tmp[indx].end;
		}
	}

	delete[] tmp;
	
	return i;	
}

/************************************************************************************/

long extractGaps(long** gaps,long len)
{
	long i,indx=0;
	long** tmp = allocateMtx(len,2);
		
	for(i=0;i<len;i++)
	{	
		tmp[i][0]=gaps[i][0];
		tmp[i][1]=gaps[i][1];
	}
	
	for(i=0;i<len;i++)
	{
		if( (unsigned long)(tmp[i][1]-tmp[i][0]+1) > gapExtractLen)
		{	
			gaps[indx][0]=tmp[i][0];
			gaps[indx++][1]=tmp[i][1];
		}
	}

	freeMtx(tmp,len);

	return indx;			
}

/************************************************************************************/

void getIntersection(long** gaps,long** gaps1,long** gaps2,long len1,long len2)
{
	long i=1;
	
	mergeGaps(gaps,gaps1,gaps2,len1,len2);

	while(i<len1+len2)
	{	
		if( gaps[i][1]<=gaps[i-1][1] )
		{
			gaps[i][1]=gaps[i-1][1];
			gaps[i][0]=gaps[i-1][0];
			gaps[i-1][2]=0;
		}
			
		else
			if( gaps[i][0]<=gaps[i-1][1] )
			{
				gaps[i][0]=gaps[i-1][0];
				gaps[i-1][2]=0;
			}
			
		i++;
	}
		
}

/************************************************************************************/

void mergeGaps(long** gaps,long** gaps1,long** gaps2,long len1,long len2)
{
	long a=0,b=0,c=0;

	while( (a<len1)&&(b<len2) )
	{
		if(gaps1[a][0]<gaps2[b][0])
		{
			gaps[c][0]=gaps1[a][0];
			gaps[c][1]=gaps1[a++][1];
			gaps[c++][2]=1;
			continue;
		}
		else	
		{	
			gaps[c][0]=gaps2[b][0];
			gaps[c][1]=gaps2[b++][1];
			gaps[c++][2]=1;
			continue;
		}
	}

	while(a<len1)
	{
		gaps[c][0]=gaps1[a][0];
		gaps[c][1]=gaps1[a++][1];
		gaps[c++][2]=1;
	}
			
	while(b<len2)
	{
		gaps[c][0]=gaps2[b][0];
		gaps[c][1]=gaps2[b++][1];
		gaps[c++][2]=1;
	}
}

/************************************************************************************/

long** allocateMtx(long len,long wid)
{
	long** mtx = new long*[len];
	if(mtx==NULL) 
	{
		cerr<<"Memory Allocation Error."<<endl;
		//errors<<"Memory Allocation Error."<<endl;
		//finish.open((input3+string("done")).c_str());
		exit(1);	
	}
	for(long j=0;j<len;j++)
	{
		mtx[j]=(long*) new long[wid];
		if(mtx[j]==NULL) 
		{
			cerr<<"Memory Allocation Error."<<endl;
			//errors<<"Memory Allocation Error."<<endl;
			//finish.open((input3+string("done")).c_str());
			exit(1);
		}
	}
	return mtx;
}

/************************************************************************************/

void freeMtx(long** mtx,long len)
{
	if (mtx == NULL) return;
	
	for(long i=0;i<len;i++)
		delete[] mtx[i];
	delete[] mtx;
}

/************************************************************************************/

