/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */


#ifndef _UMR_
#define _UMR_

#include <iostream>

using namespace std;

/*******************************************************/
/*********************** umr.h *************************/
/*******************************************************/

/**
 * node class is an auxiliary class that was used in file umr.cpp .
 */

class node
{
	public:
		long start,end;
		int valid;
};

/**
 * Allocates a matrix with "len" lines and "wid" columns.
 * @param len: long
 * @param wid: long
 * @return long**
 */
long** allocateMtx(long len,long wid);

/**
 * Frees an allocated matrix with "len" lines.
 * @param mtx: long**
 * @param len: long
 */
void freeMtx(long** mtx,long len);

/**
 * Finds the starting and ending coordinates of the gaps found in the given sequence and returns the number of them.
 * @param seq: string
 * @param gaps: long**
 * @ return long
 * Note: The matrix supplied should be with at least 2 columns (first column for the starting coordinate and the second for the ending coordinate)
 *       and long enough to hold all gaps.
 */
long findGaps(string seq,long** gaps);

/**
 * Joins gaps according to a predefined threshold (gapJoinLen), and returns the number of gaps in the given matrix after the joining .
 * @param gaps: long**
 * @param len: long
 * @return long
 * Note: "gaps" must include "len" pairs of gap coordinates.
 */  
long joinGaps(long** gaps,long len);

/**
 * Extracts gaps according to a predefined threshold (gapExtractLen), and returns the number of gaps in the given matrix after the extracting.
 * @param gaps: long**
 * @param len: long
 * @return long
 * Note: "gaps" must include "len" pairs of gap coordinates.
 */
long extractGaps(long** gaps, long len);

/**
 * Extracts unmatched regions according to MAGIC's paper.
 * @param seq1: The first sequence.
 * @param seq2: The second sequence.
 * @param gaps: The matrix where the result will be saved.
 * @return long: The number of unmatched regions found in "gaps".
 * Note: "gaps" must have enough space to hold all unmatched regions.
 */
long extractUnmatchedRegions(string seq1,string seq2,long** gaps);

/**
 * Returns the intersection between "gaps1" and "gaps2" in "gaps".
 * @param gaps: long**
 * @param gaps1: long**
 * @param gaps2: long**
 * @param len1: long
 * @param len2: long
 * Note: "len1" and "len2" are the number of lines in "gaps1" and "gaps2" respectively.
 * Important: "gaps" should include at least three columns.
 * Important: At the end of the function "gaps"'s size will be len1+len2, where the unnecessary entries will be tagged with the 
 *            value 0 in the third column (1 otherwise).
 */
void getIntersection(long** gaps,long** gaps1,long** gaps2,long len1,long len2);

/**
 * Performs a simple merging between sorted matrixes (gaps1 + gaps2) and returns the result in "gaps" matrix.
 * @param gaps: long**
 * @param gaps1: long** 
 * @param gaps2: long**
 * @param len1: long
 * @param len2: long
 * Note: "len1" and "len2" are the number of lines in "gaps1" and "gaps2" respectively.
 */
void mergeGaps(long** gaps,long** gaps1,long** gaps2,long len1,long len2);

/*******************************************************/

#endif
